@@ -1,6 +1,7 @@
 /* fish.c:
  *
  * Copyright (C) 1998-2002 Free Software Foundation, Inc.
+ * Copyright (C) 2002-2005 Vincent Untz
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -20,18 +21,23 @@
  * Authors:
  *      George Lebl  <jirka@5z.com>
  *      Mark McLoughlin <mark@skynet.ie>
+ *      Vincent Untz <vuntz@gnome.org>
  */
 
 #include <config.h>
 #include <string.h>
+#include <time.h>
 
+#include <math.h>
+#include <cairo/cairo.h>
 #include <glib-object.h>
 #include <gtk/gtk.h>
+#include <gdk/gdkkeysyms.h>
 #include <panel-applet.h>
 #include <panel-applet-gconf.h>
 #include <gconf/gconf-client.h>
-#include <libgnomeui/libgnomeui.h>
-#include <libart_lgpl/libart.h>
+#include <libgnomeui/gnome-help.h>
+#include <libgnomeui/gnome-uidefs.h>
 #include <glade/glade-xml.h>
 
 #define FISH_APPLET(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), \
@@ -42,6 +48,11 @@
 
 #define N_FISH_PREFS 6
 
+#define LOCKDOWN_COMMANDLINE_KEY "/desktop/gnome/lockdown/disable_command_line"
+#define N_FISH_LOCKDOWN 1
+
+#define N_FISH_LISTENERS (N_FISH_PREFS + N_FISH_LOCKDOWN)
+
 typedef struct {
 	PanelApplet        applet;
 
@@ -62,6 +73,7 @@ typedef struct {
 	GdkPixmap         *pixmap;
 	guint              timeout;
 	int                current_frame;
+	gboolean           in_applet;
 
 	GdkPixbuf         *pixbuf;
 
@@ -69,9 +81,10 @@ typedef struct {
 
 	GtkWidget         *preferences_dialog;
 	GtkWidget         *name_entry;
-	GtkWidget         *pixmap_entry;
-	GtkWidget         *image_entry;
+	GtkWidget         *command_label;
 	GtkWidget         *command_entry;
+	GtkWidget         *preview_image;
+	GtkWidget         *image_chooser;
 	GtkWidget         *frames_spin;
 	GtkWidget         *speed_spin;
 	GtkWidget         *rotate_toggle;
@@ -79,11 +92,15 @@ typedef struct {
 	GtkWidget         *fortune_dialog;
 	GtkWidget         *fortune_view;
 	GtkWidget         *fortune_label;
+	GtkWidget         *fortune_cmd_label;
 	GtkTextBuffer	  *fortune_buffer;
 
+	unsigned int       source_id;
+	GIOChannel        *io_channel;
+
 	gboolean           april_fools;
 
-	guint              listeners [N_FISH_PREFS];
+	guint              listeners [N_FISH_LISTENERS];
 } FishApplet;
 
 typedef struct {
@@ -95,6 +112,7 @@ static gboolean load_fish_image          (FishApplet *fish);
 static void     update_pixmap            (FishApplet *fish);
 static void     something_fishy_going_on (FishApplet *fish,
 					  const char *message);
+static void     display_fortune_dialog   (FishApplet *fish);
 static void     set_tooltip              (FishApplet *fish);
 
 static GType fish_applet_get_type (void);
@@ -106,13 +124,26 @@ static int fools_month      = 0;
 static int fools_hour_start = 0;
 static int fools_hour_end   = 0;
 
+static char *
+get_image_path (FishApplet *fish)
+{
+	char *path;
+
+	if (g_path_is_absolute (fish->image))
+		path = g_strdup (fish->image);
+	else
+		path = g_strdup_printf ("%s/%s", FISH_ICONDIR, fish->image);
+
+	return path;
+}
+
 static void
 show_help (FishApplet *fish, const char *link_id)
 {
 	GError *error = NULL;
 
 	gnome_help_display_desktop_on_screen (
-		NULL, "fish-applet-2", "fish-applet-2", link_id,
+		NULL, "fish", "fish", link_id,
 		gtk_widget_get_screen (GTK_WIDGET (fish)),
 		&error);
 
@@ -155,18 +186,36 @@ name_value_changed (GtkEntry   *entry,
 }
 
 static void
-image_value_changed (GtkEntry   *entry,
-		     FishApplet *fish)
+image_value_changed (GtkFileChooser *chooser,
+		     FishApplet     *fish)
 {	char *path;
+	char *image;
+	char *path_gconf;
 
-	path = gnome_file_entry_get_full_path (
-			GNOME_FILE_ENTRY (fish->pixmap_entry), TRUE);
-	if (!path || !path [0]) {
+	path = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (chooser));
+
+	if (!path || !path[0]) {
 		g_free (path);
 		return;
 	}
 
-	panel_applet_gconf_set_string (PANEL_APPLET (fish), "image", path, NULL);
+	path_gconf = get_image_path (fish);
+	if (!strcmp (path, path_gconf)) {
+		g_free (path);
+		g_free (path_gconf);
+		return;
+	}
+	g_free (path_gconf);
+
+	if (!strncmp (path, FISH_ICONDIR, strlen (FISH_ICONDIR))) {
+		image = path + strlen (FISH_ICONDIR);
+		while (*image && *image == G_DIR_SEPARATOR)
+			image++;
+	} else
+		image = path;
+
+	panel_applet_gconf_set_string (PANEL_APPLET (fish), "image",
+				       image, NULL);
 
 	g_free (path);
 }
@@ -304,14 +353,40 @@ setup_sensitivity (FishApplet *fish,
 
 }
 
+static void
+chooser_preview_update (GtkFileChooser *file_chooser,
+			gpointer data)
+{
+	GtkWidget *preview;
+	char      *filename;
+	GdkPixbuf *pixbuf;
+	gboolean   have_preview;
+
+	preview = GTK_WIDGET (data);
+	filename = gtk_file_chooser_get_preview_filename (file_chooser);
+
+	pixbuf = gdk_pixbuf_new_from_file_at_size (filename, 128, 128, NULL);
+	have_preview = (pixbuf != NULL);
+	g_free (filename);
+
+	gtk_image_set_from_pixbuf (GTK_IMAGE (preview), pixbuf);
+	if (pixbuf)
+		g_object_unref (pixbuf);
+
+	gtk_file_chooser_set_preview_widget_active (file_chooser,
+						    have_preview);
+}
+
 static void 
 display_preferences_dialog (BonoboUIComponent *uic,
 			    FishApplet        *fish,
 			    const char        *verbname)
 {
-	GladeXML  *xml;
-	GtkWidget *button;
-	GConfClient *client;
+	GladeXML      *xml;
+	GtkWidget     *button;
+	GtkFileFilter *filter;
+	GtkWidget     *chooser_preview;
+	char          *path;
 
 	if (fish->preferences_dialog) {
 		gtk_window_set_screen (GTK_WINDOW (fish->preferences_dialog),
@@ -331,10 +406,6 @@ display_preferences_dialog (BonoboUIComponent *uic,
 	gtk_dialog_set_default_response (
 		GTK_DIALOG (fish->preferences_dialog), GTK_RESPONSE_OK);
 
-	gnome_window_icon_set_from_file (
-		GTK_WINDOW (fish->preferences_dialog),
-		GNOME_ICONDIR "/gnome-fish.png");
-
 	fish->name_entry = glade_xml_get_widget (xml, "name_entry");
 	gtk_entry_set_text (GTK_ENTRY (fish->name_entry), fish->name);
 
@@ -347,20 +418,39 @@ display_preferences_dialog (BonoboUIComponent *uic,
 			   NULL /* label_post */,
 			   "name" /* key */);
 
-	fish->pixmap_entry = glade_xml_get_widget (xml, "image_entry");
-	fish->image_entry = gnome_file_entry_gtk_entry (
-				GNOME_FILE_ENTRY (fish->pixmap_entry));
-	gtk_entry_set_text (GTK_ENTRY (fish->image_entry), fish->image);
+	fish->preview_image = glade_xml_get_widget (xml, "preview_image");
+	if (fish->pixbuf)
+		gtk_image_set_from_pixbuf (GTK_IMAGE (fish->preview_image),
+					   fish->pixbuf);
 
-	g_signal_connect (fish->image_entry, "changed",
+	fish->image_chooser =  glade_xml_get_widget (xml, "image_chooser");
+	filter = gtk_file_filter_new ();
+	gtk_file_filter_set_name (filter, _("Images"));
+	gtk_file_filter_add_pixbuf_formats (filter);
+	gtk_file_chooser_add_filter (GTK_FILE_CHOOSER (fish->image_chooser),
+				     filter);
+	gtk_file_chooser_set_filter (GTK_FILE_CHOOSER (fish->image_chooser),
+				     filter);
+	chooser_preview = gtk_image_new ();
+	gtk_file_chooser_set_preview_widget (GTK_FILE_CHOOSER (fish->image_chooser),
+					     chooser_preview);
+	g_signal_connect (fish->image_chooser, "update-preview",
+			  G_CALLBACK (chooser_preview_update), chooser_preview);
+	path = get_image_path (fish);
+	gtk_file_chooser_set_filename (GTK_FILE_CHOOSER (fish->image_chooser),
+				       path);
+	g_free (path);
+
+	g_signal_connect (fish->image_chooser, "selection-changed",
 			  G_CALLBACK (image_value_changed), fish);
 
 	setup_sensitivity (fish, xml,
-			   "image_entry" /* wid */,
+			   "image_chooser" /* wid */,
 			   "image_label" /* label */,
 			   NULL /* label_post */,
 			   "image" /* key */);
 
+	fish->command_label = glade_xml_get_widget (xml, "command_label");
 	fish->command_entry = glade_xml_get_widget (xml, "command_entry");
 	gtk_entry_set_text (GTK_ENTRY (fish->command_entry), fish->command);
 
@@ -373,17 +463,11 @@ display_preferences_dialog (BonoboUIComponent *uic,
 			   NULL /* label_post */,
 			   "command" /* key */);
 
-	client = gconf_client_get_default ();
-	if (gconf_client_get_bool (client, "/desktop/gnome/lockdown/inhibit_command_line", NULL)) {
-		GtkWidget *w;
-
-		w = glade_xml_get_widget (xml, "command_entry");
-		g_assert (w != NULL);
-		gtk_widget_set_sensitive (w, FALSE);
-
-		w = glade_xml_get_widget (xml, "command_label");
-		g_assert (w != NULL);
-		gtk_widget_set_sensitive (w, FALSE);
+	if (gconf_client_get_bool (fish->client,
+				   LOCKDOWN_COMMANDLINE_KEY,
+				   NULL)) {
+		gtk_widget_set_sensitive (fish->command_label, FALSE);
+		gtk_widget_set_sensitive (fish->command_entry, FALSE);
 	}
 
 	fish->frames_spin = glade_xml_get_widget (xml, "frames_spin");
@@ -468,10 +552,8 @@ display_about_dialog (BonoboUIComponent *uic,
 		"Sun GNOME Documentation Team <gdocteam@sun.com>",
           	NULL
 	};
+
 	char        *authors [3];
-	GdkPixbuf   *pixbuf;
-	GError      *error = NULL;
-	char        *file;
 	char        *descr;
 
 	if (fish->about_dialog) {
@@ -485,40 +567,27 @@ display_about_dialog (BonoboUIComponent *uic,
 	authors [1] = _("(with minor help from George)");
 	authors [2] = NULL;
 
-	file = gnome_program_locate_file (NULL, GNOME_FILE_DOMAIN_PIXMAP,
-					  "gnome-fish.png", FALSE, NULL);
-	pixbuf = gdk_pixbuf_new_from_file (file, &error);
-	g_free (file);
-
-	if (error) {
-		g_warning (G_STRLOC ": cannot open %s: %s", file, error->message);
-		g_error_free (error);
-	}
-
 	descr = g_strdup_printf (about_format, fish->name);
 		
-	fish->about_dialog =
-		gnome_about_new (_("Fish"),
-				 "3.4.7.4ac19",
-				 "Copyright \xc2\xa9 1998-2002 Free Software Foundation, Inc.",
-				 descr,
-				 (const char **) authors,
-				 documenters,
-				 NULL,
-				 pixbuf);
+	fish->about_dialog = gtk_about_dialog_new ();
+	g_object_set (fish->about_dialog,
+		      "name", _("Fish"),
+		      "version", "3.4.7.4ac19",
+		      "copyright", "Copyright \xc2\xa9 1998-2002 Free Software Foundation, Inc.",
+		      "comments", descr,
+		      "authors", (const char **) authors,
+		      "documenters", documenters,
+		      "translator-credits", _("translator-credits"),
+		      "logo-icon-name", "gnome-fish",
+		      NULL);
 
 	g_free (descr);
 	g_free (authors [0]);
 
-	if (pixbuf)
-		g_object_unref (pixbuf);
-
 	gtk_window_set_wmclass (
 		GTK_WINDOW (fish->about_dialog), "fish", "Fish");
 	gtk_window_set_screen (GTK_WINDOW (fish->about_dialog),
 			       gtk_widget_get_screen (GTK_WIDGET (fish)));
-	gnome_window_icon_set_from_file (GTK_WINDOW (fish->about_dialog),
-					 GNOME_ICONDIR "/gnome-fish.png");
 
 	g_signal_connect (fish->about_dialog, "destroy",
 			  G_CALLBACK (gtk_widget_destroyed),
@@ -571,33 +640,50 @@ something_fishy_going_on (FishApplet *fish,
 	gtk_widget_show (dialog);
 }
 
-static char *
-locate_fortune_command (FishApplet *fish)
+static gboolean
+locate_fortune_command (FishApplet   *fish,
+			int          *argcp,
+			char       ***argvp)
 {
-	char *retval = NULL;
+	char *prog = NULL;
 
-	if (!fish->command) {
-		something_fishy_going_on (
-			fish, _("Unable to get the name of the command to execute"));
-		return NULL;
+	if (fish->command
+	    && g_shell_parse_argv (fish->command, argcp, argvp, NULL)) {
+		prog = g_find_program_in_path ((*argvp)[0]);
+		if (prog) {
+			g_free (prog);
+			return TRUE;
+		}
+
+		g_strfreev (*argvp);
 	}
 
-	if (g_path_is_absolute (fish->command))
-		retval = g_strdup (fish->command);
-	else
-		retval = g_find_program_in_path (fish->command);
+	prog = g_find_program_in_path ("fortune");
+	if (prog) {
+		g_free (prog);
+		if (g_shell_parse_argv ("fortune", argcp, argvp, NULL))
+			return FALSE;
+	}
 
-	if (!retval)
-		retval = g_find_program_in_path ("fortune");
+	if (g_file_test ("/usr/games/fortune", G_FILE_TEST_IS_EXECUTABLE)
+	    && g_shell_parse_argv ("/usr/games/fortune", argcp, argvp, NULL))
+		return FALSE;
 
-	if (!retval && g_file_test ("/usr/games/fortune", G_FILE_TEST_EXISTS))
-		retval = g_strdup ("/usr/games/fortune");
+	something_fishy_going_on (fish,
+				  _("Unable to locate the command to execute"));
+	*argvp = NULL;
+	return FALSE;
+}
 
-	if (!retval)
-		something_fishy_going_on (
-			fish, _("Unable to locate the command to execute"));
-
-	return retval;
+#define FISH_RESPONSE_SPEAK 1
+static inline void
+fish_close_channel (FishApplet *fish)
+{
+	if (fish->io_channel) {
+		g_io_channel_shutdown (fish->io_channel, TRUE, NULL);
+		g_io_channel_unref (fish->io_channel);
+	}
+	fish->io_channel = NULL;
 }
 
 static void
@@ -605,7 +691,17 @@ handle_fortune_response (GtkWidget  *widget,
 			 int         id,
 			 FishApplet *fish)
 {
-	gtk_widget_hide (fish->fortune_dialog);
+	if (id == FISH_RESPONSE_SPEAK)
+		display_fortune_dialog (fish);
+	else {
+		/* if there is still a pipe, close it: if we hide the widget,
+		 * the * output can't be seen */
+		if (fish->source_id)
+			g_source_remove (fish->source_id);
+		fish->source_id = 0;
+		fish_close_channel (fish);
+		gtk_widget_hide (fish->fortune_dialog);
+	}
 }
 
 static void
@@ -642,8 +738,12 @@ insert_fortune_text (FishApplet *fish,
 
 	gtk_text_buffer_get_iter_at_offset (fish->fortune_buffer, &iter, -1);
 
-	gtk_text_buffer_insert (fish->fortune_buffer, &iter,
-				text, -1);
+	gtk_text_buffer_insert_with_tags_by_name (fish->fortune_buffer, &iter,
+						  text, -1, "monospace_tag",
+						  NULL);
+
+	while (gtk_events_pending ())
+	  gtk_main_iteration ();
 }
 
 static void
@@ -655,21 +755,93 @@ clear_fortune_text (FishApplet *fish)
 	gtk_text_buffer_get_iter_at_offset (fish->fortune_buffer, &end, -1);
 
 	gtk_text_buffer_delete (fish->fortune_buffer, &begin, &end);
+	gtk_text_buffer_remove_tag_by_name (fish->fortune_buffer,
+					    "monospace_tag", &begin, &end);
 
 	/* insert an empty line */
 	insert_fortune_text (fish, "\n");
 }
 
+static gboolean
+fish_read_output (GIOChannel   *source,
+		  GIOCondition  condition,
+		  gpointer      data)
+{
+	char        output[4096];
+	char       *utf8_output;
+	gsize       bytes_read;
+	GError     *error = NULL;
+	GIOStatus   status;
+	FishApplet *fish;
+
+	fish = (FishApplet *) data;
+
+	if (!(condition & G_IO_IN)) {
+		fish->source_id = 0;
+		fish_close_channel (fish);
+		return FALSE;
+	}
+
+	status = g_io_channel_read_chars (source, output, 4096, &bytes_read,
+					  &error);
+
+	if (error) {
+		char *message;
+
+		message = g_strdup_printf (_("Unable to read output from command\n\nDetails: %s"),
+					   error->message);
+		something_fishy_going_on (fish, message);
+		g_free (message);
+		g_error_free (error);
+		fish->source_id = 0;
+		fish_close_channel (fish);
+		return FALSE;
+	}
+
+	if (status == G_IO_STATUS_AGAIN)
+		return TRUE;
+
+	if (bytes_read > 0) {
+		/* The output is not guarantied to be in UTF-8 format, most
+		 * likely it's just in ASCII-7 or in the user locale
+		 */
+		if (!g_utf8_validate (output, -1, NULL))
+			utf8_output = g_locale_to_utf8 (output, bytes_read,
+							NULL, NULL, NULL);
+		else
+			utf8_output = g_strndup (output, bytes_read);
+
+		if (utf8_output)
+			insert_fortune_text (fish, utf8_output);
+
+		g_free (utf8_output);
+	}
+
+	if (status == G_IO_STATUS_EOF) {
+		fish->source_id = 0;
+		fish_close_channel (fish);
+	}
+	return (status != G_IO_STATUS_EOF);
+}
+
 static void 
 display_fortune_dialog (FishApplet *fish)
 {
-	GError *error = NULL;
-	char   *fortune_command;
-	char   *output = NULL;
-	char   *utf8_output;
+	GError      *error = NULL;
+	gboolean     user_command;
+	int          output;
+	const char  *charset;
+	int          argc;
+	char       **argv;
 
-	fortune_command = locate_fortune_command (fish);
-	if (!fortune_command)
+	/* if there is still a pipe, close it */
+	if (fish->source_id)
+		g_source_remove (fish->source_id);
+	fish->source_id = 0;
+	fish_close_channel (fish);
+
+	user_command = locate_fortune_command (fish, &argc, &argv);
+	if (!argv)
 		return;
 
 	if (!fish->fortune_dialog) {
@@ -681,6 +853,7 @@ display_fortune_dialog (FishApplet *fish)
 		fish->fortune_dialog = 
 			gtk_dialog_new_with_buttons (
 				"", NULL, 0,
+				_("_Speak again"), FISH_RESPONSE_SPEAK,
 				GTK_STOCK_CLOSE, GTK_RESPONSE_CLOSE,
 				NULL);
 
@@ -696,16 +869,12 @@ display_fortune_dialog (FishApplet *fish)
 				  G_CALLBACK (handle_fortune_response), fish);
 
 		gtk_window_set_wmclass (GTK_WINDOW (fish->fortune_dialog), "fish", "Fish");
-		gnome_window_icon_set_from_file (GTK_WINDOW (fish->fortune_dialog),
-						 GNOME_ICONDIR"/gnome-fish.png");
 
 		screen = gtk_widget_get_screen (GTK_WIDGET (fish));
 
 		screen_width  = gdk_screen_get_width (screen);
 		screen_height = gdk_screen_get_height (screen);
 
-		gtk_window_set_screen (GTK_WINDOW (fish->fortune_dialog), screen);
-
 		gtk_window_set_default_size (GTK_WINDOW (fish->fortune_dialog),
 					     MIN (600, screen_width  * 0.9),
 					     MIN (350, screen_height * 0.9));
@@ -718,6 +887,10 @@ display_fortune_dialog (FishApplet *fish)
 		fish->fortune_buffer =
 			gtk_text_view_get_buffer (GTK_TEXT_VIEW (fish->fortune_view));
 
+		gtk_text_buffer_create_tag (GTK_TEXT_BUFFER (fish->fortune_buffer),
+					    "monospace_tag", "family",
+					    "Monospace", NULL);
+
 		scrolled = gtk_scrolled_window_new (NULL, NULL);
 		gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolled),
 						GTK_POLICY_AUTOMATIC,
@@ -728,6 +901,11 @@ display_fortune_dialog (FishApplet *fish)
 		gtk_container_add (GTK_CONTAINER (scrolled), fish->fortune_view);
 
 		fish->fortune_label = gtk_label_new ("");
+		gtk_label_set_ellipsize (GTK_LABEL (fish->fortune_label),
+					 PANGO_ELLIPSIZE_MIDDLE);
+		fish->fortune_cmd_label = gtk_label_new ("");
+		gtk_misc_set_alignment (GTK_MISC (fish->fortune_cmd_label),
+					0, 0.5);
 
 		gtk_box_pack_start (GTK_BOX (GTK_DIALOG (fish->fortune_dialog)->vbox), 
 				    fish->fortune_label,
@@ -737,49 +915,82 @@ display_fortune_dialog (FishApplet *fish)
 				    scrolled,
 				    TRUE, TRUE, GNOME_PAD);
 
+		gtk_box_pack_start (GTK_BOX (GTK_DIALOG (fish->fortune_dialog)->vbox), 
+				    fish->fortune_cmd_label,
+				    FALSE, FALSE, GNOME_PAD);
+
 		update_fortune_dialog (fish);
 
-		gtk_widget_show_all (fish->fortune_dialog);
+		/* We don't show_all for the dialog since fortune_cmd_label
+		 * might need to be hidden 
+		 * The dialog will be shown with gtk_window_present later */
+		gtk_widget_show (scrolled);
+		gtk_widget_show (fish->fortune_view);
+		gtk_widget_show (fish->fortune_label);
 	}
 
-	gtk_window_set_screen (GTK_WINDOW (fish->fortune_dialog),
-			       gtk_widget_get_screen (GTK_WIDGET (fish)));
-	gtk_window_present (GTK_WINDOW (fish->fortune_dialog));
+	if (!user_command) {
+		char *command;
+		char * text;
+
+		command = g_strdup_printf ("<tt>%s</tt>", argv[0]);
+		text = g_strdup_printf (_("The configured command is not "
+					  "working and has been replaced by: "
+					  "%s"), command);
+		gtk_label_set_markup (GTK_LABEL (fish->fortune_cmd_label),
+				      text);
+		g_free (command);
+		g_free (text);
+		gtk_widget_show (fish->fortune_cmd_label);
+	} else {
+		gtk_widget_hide (fish->fortune_cmd_label);
+	}
 
 	clear_fortune_text (fish);
 
-	g_spawn_command_line_sync (fortune_command, &output, NULL, NULL, &error);
+	gdk_spawn_on_screen_with_pipes (gtk_widget_get_screen (GTK_WIDGET (fish)),
+					NULL, argv, NULL,
+					G_SPAWN_SEARCH_PATH|G_SPAWN_STDERR_TO_DEV_NULL,
+					NULL, NULL, NULL, NULL, &output, NULL,
+					&error);
+
 	if (error) {
 		char *message;
 
 		message = g_strdup_printf (_("Unable to execute '%s'\n\nDetails: %s"),
-					   fortune_command, error->message);
+					   argv[0], error->message);
 		something_fishy_going_on (fish, message);
 		g_free (message);
 		g_error_free (error);
+		g_strfreev (argv);
+		return;
 	}
-	
-	g_free (fortune_command);
 
-	/* The output is not guarantied to be in UTF-8 format, most
-	 * likely it's just in ASCII-7 or in the user locale
-	  */
-	if (!g_utf8_validate (output, -1, NULL))
-		utf8_output = g_locale_to_utf8 (output, -1, NULL, NULL, NULL);
-	else
-		utf8_output = g_strdup (output);
+	fish->io_channel = g_io_channel_unix_new (output);
+	/* set the correct encoding if the locale is not using UTF-8 */
+	if (!g_get_charset (&charset))
+		g_io_channel_set_encoding(fish->io_channel, charset, &error);
+	if (error) {
+		char *message;
 
-	if (utf8_output)
-		insert_fortune_text (fish, utf8_output);
-	else
-		insert_fortune_text (fish, 
-				     _("You do not have fortune installed "
-				       "or you have not specified a program "
-				       "to run.\n\nPlease refer to fish "
-				       "properties dialog."));
+		message = g_strdup_printf (_("Unable to read from '%s'\n\nDetails: %s"),
+					   argv[0], error->message);
+		something_fishy_going_on (fish, message);
+		g_free (message);
+		g_error_free (error);
+		g_strfreev (argv);
+		return;
+	}
 
-	g_free (output);
-	g_free (utf8_output);
+	g_strfreev (argv);
+
+	fish->source_id = g_io_add_watch (fish->io_channel,
+					  G_IO_IN|G_IO_ERR|G_IO_HUP|G_IO_NVAL,
+					  fish_read_output, fish);
+
+	gtk_window_set_screen (GTK_WINDOW (fish->fortune_dialog),
+			       gtk_widget_get_screen (GTK_WIDGET (fish)));
+	gtk_window_present (GTK_WINDOW (fish->fortune_dialog));
 }
 
 static void
@@ -834,9 +1045,19 @@ image_changed_notify (GConfClient *client,
 	load_fish_image (fish);
 	update_pixmap (fish);
 
-	if (fish->image_entry &&
-	    strcmp (gtk_entry_get_text (GTK_ENTRY (fish->image_entry)), fish->image))
-		gtk_entry_set_text (GTK_ENTRY (fish->image_entry), fish->image);
+	if (fish->image_chooser) {
+		char *path_gconf;
+		char *path_chooser;
+
+		path_gconf = get_image_path (fish);
+		path_chooser = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (fish->image_chooser));
+		if (strcmp (path_gconf, path_chooser))
+			gtk_file_chooser_set_filename (GTK_FILE_CHOOSER (fish->image_chooser),
+						       path_gconf);
+
+		g_free (path_gconf);
+		g_free (path_chooser);
+	}
 }
 
 static void
@@ -852,7 +1073,7 @@ command_changed_notify (GConfClient *client,
 	
 	value = gconf_value_get_string (entry->value);
 
-	if (!value [0] || (fish->command && !strcmp (fish->command, value)))
+	if (fish->command && !strcmp (fish->command, value))
 		return;
 
 	if (fish->command)
@@ -1065,6 +1286,25 @@ rotate_changed_notify (GConfClient *client,
 			GTK_TOGGLE_BUTTON (fish->rotate_toggle), fish->rotate);
 }
 
+static void
+fish_disable_commande_line_notify (GConfClient *client,
+				   guint        cnxn_id,
+				   GConfEntry  *entry,
+				   FishApplet  *fish)
+{
+	gboolean locked_down;
+
+	if (!entry->value || entry->value->type != GCONF_VALUE_BOOL)
+		return;
+
+	locked_down = !gconf_value_get_bool (entry->value);
+
+	if (fish->command_label != NULL)
+		gtk_widget_set_sensitive (fish->command_label, locked_down);
+	if (fish->command_entry != NULL)
+		gtk_widget_set_sensitive (fish->command_entry, locked_down);
+}
+
 static void
 setup_gconf (FishApplet *fish)
 {
@@ -1114,7 +1354,13 @@ setup_gconf (FishApplet *fish)
 					fish, NULL, NULL);
 	g_free (key);
 
-	g_assert (i == N_FISH_PREFS);
+	fish->listeners [i++] = gconf_client_notify_add (
+					fish->client,
+					LOCKDOWN_COMMANDLINE_KEY,
+					(GConfClientNotifyFunc) fish_disable_commande_line_notify,
+					fish, NULL, NULL);
+
+	g_assert (i == N_FISH_LISTENERS);
 }
 
 static gboolean
@@ -1127,23 +1373,7 @@ load_fish_image (FishApplet *fish)
 	if (!fish->image)
 		return FALSE;
 
-	if (g_path_is_absolute (fish->image))
-		path = g_strdup (fish->image);
-
-	else {
-		path = gnome_program_locate_file (
-					NULL, GNOME_FILE_DOMAIN_PIXMAP,
-					fish->image, FALSE, NULL);
-		if (!path)
-			path = gnome_program_locate_file (
-					NULL, GNOME_FILE_DOMAIN_APP_PIXMAP,
-					fish->image, FALSE, NULL);
-	}
-
-	if (!path) {
-		g_warning ("Cannot locate '%s'", fish->image);
-		return FALSE;
-	}
+	path = get_image_path (fish);
 
 	pixbuf = gdk_pixbuf_new_from_file (path, &error);
 	if (error) {
@@ -1157,6 +1387,10 @@ load_fish_image (FishApplet *fish)
 		g_object_unref (fish->pixbuf);
 	fish->pixbuf = pixbuf;
 
+	if (fish->preview_image)
+		gtk_image_set_from_pixbuf (GTK_IMAGE (fish->preview_image),
+					   fish->pixbuf);
+
 	g_free (path);
 
 	return TRUE;
@@ -1166,52 +1400,45 @@ static void
 update_pixmap (FishApplet *fish)
 {
 	GtkWidget *widget = fish->drawing_area;
-	GdkGC     *gc;
 	int        width  = -1;
 	int        height = -1;
 	int        pixbuf_width = -1;
 	int        pixbuf_height = -1;
 	gboolean   rotate = FALSE;
-	double     affine [6];
-	guchar    *rgb;
+	cairo_t   *cr;
+	cairo_matrix_t matrix;
+	cairo_pattern_t *pattern;
 
 	if (!GTK_WIDGET_REALIZED (widget) ||
 	    widget->allocation.width <= 0 ||
 	    widget->allocation.height <= 0)
 		return;
 
+	if (!fish->pixbuf && !load_fish_image (fish))
+		return;
+
 	if (fish->rotate &&
 	    (fish->orientation == PANEL_APPLET_ORIENT_LEFT ||
 	     fish->orientation == PANEL_APPLET_ORIENT_RIGHT))
 		rotate = TRUE;
 
-	if (fish->pixbuf || load_fish_image (fish)) {
-		pixbuf_width  = gdk_pixbuf_get_width  (fish->pixbuf);
-		pixbuf_height = gdk_pixbuf_get_height (fish->pixbuf);
+	pixbuf_width  = gdk_pixbuf_get_width  (fish->pixbuf);
+	pixbuf_height = gdk_pixbuf_get_height (fish->pixbuf);
 
-		if (fish->orientation == PANEL_APPLET_ORIENT_UP ||
-		    fish->orientation == PANEL_APPLET_ORIENT_DOWN) {
-			height = widget->allocation.height;
-			width  = pixbuf_width * ((gdouble) height / pixbuf_height);
-			widget->requisition.width = width / fish->n_frames;
-		} else {
-			if (!rotate) {
-				width = widget->allocation.width * fish->n_frames;
-				height = pixbuf_height * ((gdouble) width / pixbuf_width);
-				widget->requisition.height = height;
-			} else {
-				width = widget->allocation.width;
-				height = pixbuf_width * ((gdouble) width / pixbuf_height);
-				widget->requisition.height = height / fish->n_frames;
-			}
-		}
+	if (fish->orientation == PANEL_APPLET_ORIENT_UP ||
+	    fish->orientation == PANEL_APPLET_ORIENT_DOWN) {
+		height = widget->allocation.height;
+		width  = pixbuf_width * ((gdouble) height / pixbuf_height);
+		widget->requisition.width = width / fish->n_frames;
 	} else {
-		if (rotate) {
-			width  = widget->allocation.width;
-			height = widget->allocation.height * fish->n_frames;
+		if (!rotate) {
+			width = widget->allocation.width * fish->n_frames;
+			height = pixbuf_height * ((gdouble) width / pixbuf_width);
+			widget->requisition.height = height;
 		} else {
-			height = widget->allocation.height;
-			width  = widget->allocation.width * fish->n_frames;
+			width = widget->allocation.width;
+			height = pixbuf_width * ((gdouble) width / pixbuf_height);
+			widget->requisition.height = height / fish->n_frames;
 		}
 	}
 
@@ -1224,71 +1451,55 @@ update_pixmap (FishApplet *fish)
 		g_object_unref (fish->pixmap);
 	fish->pixmap = gdk_pixmap_new (widget->window, width, height, -1);
 
-	if (!fish->pixbuf)
-		return;
-
 	gtk_widget_queue_resize (widget);
 
 	g_assert (pixbuf_width != -1 && pixbuf_height != -1);
 
-	affine [1] = affine [2] = affine [4] = affine [5] = 0;
+	cr = gdk_cairo_create (fish->pixmap);
 
-	if (!rotate) {
-		affine [0] = width  / (double) pixbuf_width;
-		affine [3] = height / (double) pixbuf_height;
-	} else {
-		double tmp [6];
+	cairo_set_source_rgb (cr, 1, 1, 1);
+	cairo_paint (cr);
 
-		affine [0] = height / (double) pixbuf_width;
-		affine [3] = width  / (double) pixbuf_height;
-		
-		art_affine_rotate (tmp, 270);
-		art_affine_multiply (affine, affine, tmp);
-                art_affine_translate (tmp, 0, height);
-                art_affine_multiply (affine, affine, tmp);
-	}
+	gdk_cairo_set_source_pixbuf (cr, fish->pixbuf, 0, 0);
+	pattern = cairo_get_source (cr);
+	cairo_pattern_set_filter (pattern, CAIRO_FILTER_BEST);
+
+	cairo_matrix_init_identity (&matrix);
 
 	if (fish->april_fools) {
-                double tmp [6];
+		cairo_matrix_translate (&matrix,
+					pixbuf_width - 1, pixbuf_height - 1);
+		cairo_matrix_rotate (&matrix, M_PI);
+	}
 
-                art_affine_rotate (tmp, 180);
-                art_affine_multiply (affine, affine, tmp);
-                art_affine_translate (tmp, width, height);
-                art_affine_multiply (affine, affine, tmp);
-        }
+	if (rotate) {
+		if (fish->orientation == PANEL_APPLET_ORIENT_LEFT) {
+			cairo_matrix_translate (&matrix, pixbuf_width - 1, 0);
+			cairo_matrix_rotate (&matrix, M_PI * 0.5);
+		} else {
+			cairo_matrix_translate (&matrix, 0, pixbuf_height - 1);
+			cairo_matrix_rotate (&matrix, M_PI * 1.5);
+		}
+		cairo_matrix_scale (&matrix,
+				    (double) (pixbuf_height - 1) / width,
+				    (double) (pixbuf_width - 1) / height);
+	} else {
+		cairo_matrix_scale (&matrix,
+				    (double) (pixbuf_width - 1) / width,
+				    (double) (pixbuf_height - 1) / height);
+	}
 
-	rgb = g_new0 (guchar, width * height * 3);
+	cairo_pattern_set_matrix (pattern, &matrix);
 
-	if (gdk_pixbuf_get_has_alpha (fish->pixbuf))
-		art_rgb_rgba_affine (rgb, 0, 0, width, height, width * 3,
-				     gdk_pixbuf_get_pixels (fish->pixbuf),
-                                     pixbuf_width, pixbuf_height,
-                                     gdk_pixbuf_get_rowstride (fish->pixbuf),
-                                     affine,
-                                     ART_FILTER_NEAREST,
-                                     NULL);
-        else
-                art_rgb_affine (rgb, 0, 0, width, height, width * 3,
-                                gdk_pixbuf_get_pixels (fish->pixbuf),
-                                pixbuf_width, pixbuf_height,
-                                gdk_pixbuf_get_rowstride (fish->pixbuf),
-                                affine,
-                                ART_FILTER_NEAREST,
-                                NULL);
+	cairo_rectangle (cr, 0, 0, width, height);
+	cairo_fill (cr);
 
+	if (fish->april_fools) {
+		cairo_set_source_rgb (cr, 1, 0.5, 0);
+		cairo_paint_with_alpha (cr, 0.25);
+	}
 
-	if (fish->april_fools)
-                art_rgb_run_alpha (rgb, 255, 128, 0, 70, width * height);
-
-	gc = gdk_gc_new (fish->pixmap);
-
-	gdk_draw_rgb_image (fish->pixmap, gc,
-			    0, 0, width, height,
-                            GDK_RGB_DITHER_NORMAL,
-                            rgb, width * 3);
-
-	g_object_unref (gc);
-        g_free (rgb);
+	cairo_destroy (cr);
 }
 
 static gboolean
@@ -1308,11 +1519,14 @@ fish_applet_expose_event (GtkWidget      *widget,
 	src_x = event->area.x;
 	src_y = event->area.y;
 
-	if (fish->rotate &&
-	    (fish->orientation == PANEL_APPLET_ORIENT_LEFT ||
-	     fish->orientation == PANEL_APPLET_ORIENT_RIGHT))
-		src_y += ((height * fish->current_frame) / fish->n_frames);
-	else
+	if (fish->rotate) {
+		if (fish->orientation == PANEL_APPLET_ORIENT_LEFT)
+			src_y += ((height * (fish->n_frames - 1 - fish->current_frame)) / fish->n_frames);
+		else if (fish->orientation == PANEL_APPLET_ORIENT_RIGHT)
+			src_y += ((height * fish->current_frame) / fish->n_frames);
+		else
+			src_x += ((width * fish->current_frame) / fish->n_frames);
+	} else
 		src_x += ((width * fish->current_frame) / fish->n_frames);
 
 	gdk_draw_drawable (widget->window,
@@ -1369,6 +1583,42 @@ fish_applet_change_orient (PanelApplet       *applet,
 		update_pixmap (fish);
 }
 
+static void
+fish_applet_change_background (PanelApplet               *panel_applet,
+			       PanelAppletBackgroundType  type,
+			       GdkColor                  *color,
+			       GdkPixmap                 *pixmap)
+{
+	FishApplet *applet;
+	GtkRcStyle *rc_style;
+	GtkStyle   *style;
+
+	applet = (FishApplet *) panel_applet;
+
+	/* reset style */
+	gtk_widget_set_style (GTK_WIDGET (applet), NULL);
+	rc_style = gtk_rc_style_new ();
+	gtk_widget_modify_style (GTK_WIDGET (applet), rc_style);
+	gtk_rc_style_unref (rc_style);
+
+	switch (type) {
+	case PANEL_NO_BACKGROUND:
+		break;
+	case PANEL_COLOR_BACKGROUND:
+		gtk_widget_modify_bg (GTK_WIDGET (applet),
+				      GTK_STATE_NORMAL, color);
+		break;
+	case PANEL_PIXMAP_BACKGROUND:
+		style = gtk_style_copy (GTK_WIDGET (applet)->style);
+		if (style->bg_pixmap[GTK_STATE_NORMAL])
+			g_object_unref (style->bg_pixmap[GTK_STATE_NORMAL]);
+		style->bg_pixmap[GTK_STATE_NORMAL] = g_object_ref (pixmap);
+		gtk_widget_set_style (GTK_WIDGET (applet), style);
+		g_object_unref (style);
+		break;
+	}
+}
+
 static void
 change_water (FishApplet *fish)
 {
@@ -1416,11 +1666,45 @@ handle_keypress (GtkWidget   *widget,
 	return TRUE;
 }
 
-static gboolean 
-handle_button_press (FishApplet     *fish,
-		     GdkEventButton *event)
+static gboolean
+fish_enter_notify (GtkWidget        *widget,
+		   GdkEventCrossing *event)
 {
-	if (event->button != 1)
+  FishApplet *fish;
+  GtkWidget  *event_widget;
+
+  fish = FISH_APPLET (widget);
+  event_widget = gtk_get_event_widget ((GdkEvent*) event);
+
+  if ((event_widget == widget) &&
+      (event->detail != GDK_NOTIFY_INFERIOR))
+      fish->in_applet = TRUE;
+
+  return FALSE;
+}
+
+static gboolean
+fish_leave_notify (GtkWidget        *widget,
+		   GdkEventCrossing *event)
+{
+  FishApplet *fish;
+  GtkWidget  *event_widget;
+
+  fish = FISH_APPLET (widget);
+  event_widget = gtk_get_event_widget ((GdkEvent*) event);
+
+  if ((event_widget == widget) &&
+      (event->detail != GDK_NOTIFY_INFERIOR))
+      fish->in_applet = FALSE;
+
+  return FALSE;
+}
+
+static gboolean 
+handle_button_release (FishApplet     *fish,
+		       GdkEventButton *event)
+{
+	if (!fish->in_applet || event->button != 1)
 		return FALSE; 
 
 	if (fish->april_fools) {
@@ -1440,11 +1724,14 @@ set_tooltip (FishApplet *fish)
 	const char  *desc_format = _("%s the Fish, the fortune teller");
 	char        *desc;
 
-	tooltips = gtk_tooltips_new ();
-	g_object_ref (tooltips);
-	gtk_object_sink (GTK_OBJECT (tooltips));
+	tooltips = g_object_get_data (G_OBJECT (fish), "tooltips");
+	if (!tooltips) {
+		tooltips = gtk_tooltips_new ();
+		g_object_ref (tooltips);
+		gtk_object_sink (GTK_OBJECT (tooltips));
 
-	g_object_set_data (G_OBJECT (fish), "tooltips", tooltips);
+		g_object_set_data (G_OBJECT (fish), "tooltips", tooltips);
+	}
 
 	desc = g_strdup_printf (desc_format, fish->name);
 	gtk_tooltips_set_tip (tooltips, GTK_WIDGET (fish), desc, NULL);
@@ -1484,13 +1771,20 @@ setup_fish_widget (FishApplet *fish)
 	g_signal_connect (fish->drawing_area, "expose-event",
 			  G_CALLBACK (fish_applet_expose_event), fish);
 
-	gtk_widget_add_events (widget, GDK_BUTTON_PRESS_MASK);
-	g_signal_connect_swapped (widget, "button_press_event",
-				  G_CALLBACK (handle_button_press), fish);
+	gtk_widget_add_events (widget, GDK_ENTER_NOTIFY_MASK | 
+				       GDK_LEAVE_NOTIFY_MASK |
+				       GDK_BUTTON_RELEASE_MASK);
 
-	gtk_widget_add_events (fish->drawing_area, GDK_BUTTON_PRESS_MASK);
-	g_signal_connect_swapped (fish->drawing_area, "button_press_event",
-				  G_CALLBACK (handle_button_press), fish);
+	g_signal_connect_swapped (widget, "enter_notify_event",
+				  G_CALLBACK (fish_enter_notify), fish);
+	g_signal_connect_swapped (widget, "leave_notify_event",
+				  G_CALLBACK (fish_leave_notify), fish);
+	g_signal_connect_swapped (widget, "button_release_event",
+				  G_CALLBACK (handle_button_release), fish);
+
+	gtk_widget_add_events (fish->drawing_area, GDK_BUTTON_RELEASE_MASK);
+	g_signal_connect_swapped (fish->drawing_area, "button_release_event",
+				  G_CALLBACK (handle_button_release), fish);
 
 	load_fish_image (fish);
 
@@ -1515,6 +1809,22 @@ static const BonoboUIVerb fish_menu_verbs [] = {
         BONOBO_UI_VERB_END
 };
 
+static void
+fish_migrate_to_210 (FishApplet *fish)
+{
+	char *new_image;
+
+	g_assert (fish->image);
+
+	if (!strncmp (fish->image, "fish/", 5)) {
+		new_image = g_strdup (fish->image + 5);
+		g_free (fish->image);
+		fish->image = new_image;
+		panel_applet_gconf_set_string (PANEL_APPLET (fish), "image",
+					       fish->image, NULL);
+	}
+}
+
 static gboolean
 fish_applet_fill (FishApplet *fish)
 {
@@ -1544,7 +1854,8 @@ fish_applet_fill (FishApplet *fish)
 		error = NULL;
 	}
 	if (!fish->image)
-		fish->image = g_strdup ("fish/fishanim.png"); /* Fallback */
+		fish->image = g_strdup ("fishanim.png"); /* Fallback */
+	fish_migrate_to_210 (fish);
 
 	fish->command = panel_applet_gconf_get_string (applet, "command", &error);
 	if (error) {
@@ -1552,8 +1863,6 @@ fish_applet_fill (FishApplet *fish)
 		g_error_free (error);
 		error = NULL;
 	}
-	if (!fish->command)
-		fish->command = g_strdup ("fortune"); /* Fallback */
 
 	fish->n_frames = panel_applet_gconf_get_int (applet, "frames", &error);
 	if (error) {
@@ -1599,6 +1908,7 @@ fish_applet_fill (FishApplet *fish)
 					      NULL);
 	}
 
+	gtk_window_set_default_icon_name ("gnome-fish");
 	setup_fish_widget (fish);
 
 	return TRUE;
@@ -1627,8 +1937,8 @@ fish_applet_destroy (GtkObject *object)
 		g_source_remove (fish->timeout);
 	fish->timeout = 0;
 
-	for (i = 0; i < N_FISH_PREFS; i++) {
-		if (fish->client)
+	for (i = 0; i < N_FISH_LISTENERS; i++) {
+		if (fish->client && fish->listeners [i] != 0)
 			gconf_client_notify_remove (
 				fish->client, fish->listeners [i]);
 		fish->listeners [i] = 0;
@@ -1670,6 +1980,12 @@ fish_applet_destroy (GtkObject *object)
 		gtk_widget_destroy (fish->fortune_dialog);
 	fish->fortune_dialog = NULL;
 
+	if (fish->source_id)
+		g_source_remove (fish->source_id);
+	fish->source_id = 0;
+
+	fish_close_channel (fish);
+
 	destroy_tooltip (fish);
 
 	GTK_OBJECT_CLASS (parent_class)->destroy (object);
@@ -1697,6 +2013,7 @@ fish_applet_instance_init (FishApplet      *fish,
 	fish->pixmap        = NULL;
 	fish->timeout       = 0;
 	fish->current_frame = 0;
+	fish->in_applet     = FALSE;
 
 	fish->prev_allocation.x      = -1;
 	fish->prev_allocation.y      = -1;
@@ -1709,9 +2026,10 @@ fish_applet_instance_init (FishApplet      *fish,
 
 	fish->preferences_dialog = NULL;
 	fish->name_entry         = NULL;
-	fish->pixmap_entry       = NULL;
-	fish->image_entry        = NULL;
+	fish->command_label      = NULL;
 	fish->command_entry      = NULL;
+	fish->preview_image      = NULL;
+	fish->image_chooser      = NULL;
 	fish->frames_spin        = NULL;
 	fish->speed_spin         = NULL;
 	fish->rotate_toggle      = NULL;
@@ -1719,9 +2037,13 @@ fish_applet_instance_init (FishApplet      *fish,
 	fish->fortune_dialog = NULL;
 	fish->fortune_view   = NULL;
 	fish->fortune_label  = NULL;
+	fish->fortune_cmd_label = NULL;
 	fish->fortune_buffer = NULL;
 
-	for (i = 0; i < N_FISH_PREFS; i++)
+	fish->source_id  = 0;
+	fish->io_channel = NULL;
+
+	for (i = 0; i < N_FISH_LISTENERS; i++)
 		fish->listeners [i] = 0;
 
 	fish->april_fools = FALSE;
@@ -1739,6 +2061,7 @@ fish_applet_class_init (FishAppletClass *klass)
 	parent_class = g_type_class_peek_parent (klass);
 
 	applet_class->change_orient = fish_applet_change_orient;
+	applet_class->change_background = fish_applet_change_background;
 
 	gtkobject_class->destroy = fish_applet_destroy;
 
