@@ -38,6 +38,7 @@
 #include "panel-globals.h"
 #include "panel-multiscreen.h"
 #include "panel-lockdown.h"
+#include "panel-compatibility.h"
 
 #undef LAUNCHER_DEBUG
 
@@ -112,6 +113,7 @@ panel_launcher_get_filename (const char *path)
 
 static void
 panel_launcher_save_ditem (GnomeDesktopItem *ditem,
+			   gboolean          report_errors,
 			   GdkScreen        *screen)
 {
 	GError     *error = NULL;
@@ -126,15 +128,22 @@ panel_launcher_save_ditem (GnomeDesktopItem *ditem,
 		uri = panel_make_unique_uri (NULL, ".desktop");
 		gnome_desktop_item_set_location (ditem, uri);
 		g_free (uri);
+	} else if (!strchr (location, G_DIR_SEPARATOR)) {
+		char *path;
+
+		path = panel_make_full_path (NULL, location);
+		gnome_desktop_item_set_location (ditem, path);
+		g_free (path);
 	}
 
 	gnome_desktop_item_save (ditem, NULL, TRUE, &error);
 	if (error) {
-		panel_error_dialog (screen,
-				    "cannot_save_launcher", TRUE,
-				    _("Cannot save launcher to disk"),
-				    "%s",
-				    error->message);
+		if (report_errors)
+			panel_error_dialog (screen,
+					    "cannot_save_launcher", TRUE,
+					    _("Cannot save launcher to disk"),
+					    "%s",
+					    error->message);
 
 		g_error_free (error);
 	}
@@ -319,9 +328,10 @@ panel_launcher_delete (Launcher *launcher)
 {
 	const char *location;
 
-	location = gnome_desktop_item_get_location (launcher->ditem);
+	if (!(location = gnome_desktop_item_get_location (launcher->ditem)))
+		return;
 
-	if (location) {
+	if (panel_launcher_get_filename (location)) {
 		GnomeVFSResult result;
 
 		result = gnome_vfs_unlink (location);
@@ -452,7 +462,8 @@ drag_data_get_cb (GtkWidget          *widget,
 	} else if (info == TARGET_ICON_INTERNAL)
 		gtk_selection_data_set (selection_data,
 					selection_data->target, 8,
-					location, strlen (location));
+					(unsigned char *) location,
+					strlen (location));
 }
 
 static Launcher *
@@ -533,32 +544,22 @@ create_launcher (const char *location)
 }
 
 static char *
-guess_icon_from_exec (GnomeDesktopItem *ditem)
+guess_icon_from_exec (GtkIconTheme     *icon_theme,
+		      GnomeDesktopItem *ditem)
 {
 	const char *exec;
 	char       *icon_name;
-	char       *icon;
+	char       *retval;
 
 	exec = gnome_desktop_item_get_string (ditem, GNOME_DESKTOP_ITEM_EXEC);
 	if (!exec || !exec [0])
 		return NULL;
 
 	icon_name = g_path_get_basename (exec);
+	retval = panel_find_icon (icon_theme, icon_name, 48);
+	g_free (icon_name);
 
-	if (!(icon = gnome_desktop_item_find_icon (panel_icon_theme, icon_name, 48, 0))) {
-		g_free (icon_name);
-		return NULL;
-	}
-
-	if (g_file_test (icon, G_FILE_TEST_IS_DIR)) {
-		g_free (icon_name);
-		g_free (icon);
-		return NULL;
-	}
-
-	g_free (icon);
-
-	return icon_name;
+	return retval;
 }
 
 static void
@@ -584,32 +585,68 @@ setup_button (Launcher *launcher)
 	else
 		str = g_strdup (name);
 
+	/* If we can unescape the string, then we probably have an escaped
+	 * string (a location e.g.). If we can't, then it most probably means
+	 * we have a % that is not here to encode a character, and we don't
+	 * want to unescape in this case. See bug #170516 for details. */
 	unescaped_str = gnome_vfs_unescape_string (str, NULL);
-	gtk_tooltips_set_tip (panel_tooltips, launcher->button,
-			      unescaped_str, NULL);
+	if (unescaped_str) {
+		g_free (str);
+		str = unescaped_str;
+	}
 
+	gtk_tooltips_set_tip (panel_tooltips, launcher->button, str, NULL);
 	/* Setup accessible name */
-	panel_a11y_set_atk_name_desc (launcher->button, unescaped_str, NULL);
+	panel_a11y_set_atk_name_desc (launcher->button, str, NULL);
 
 	g_free (str);
-	g_free (unescaped_str);
 
 	/* Setup icon */
 	icon = gnome_desktop_item_get_string (launcher->ditem,
 					      GNOME_DESKTOP_ITEM_ICON);
 	if (!icon)
-		icon = freeme = guess_icon_from_exec (launcher->ditem);
+		icon = freeme = guess_icon_from_exec (BUTTON_WIDGET (launcher->button)->icon_theme,
+						      launcher->ditem);
 
 	button_widget_set_icon_name (BUTTON_WIDGET (launcher->button), sure_string (icon));
 
 	g_free (freeme);
 }
 
+static char *
+panel_launcher_hoard (const char *launcher_location)
+{
+	GnomeDesktopItem *ditem;
+	const char       *path;
+	char             *retval;
+
+	retval = NULL;
+	if (!launcher_location ||
+	    !strchr (launcher_location, G_DIR_SEPARATOR) ||
+	    (retval = g_strdup (panel_launcher_get_filename (launcher_location))))
+		return retval;
+
+	ditem = gnome_desktop_item_new_from_uri (launcher_location, 0, NULL);
+	if (!ditem)
+		return NULL;
+
+	gnome_desktop_item_set_location (ditem, NULL);
+	panel_launcher_save_ditem (ditem, FALSE, NULL);
+	
+	retval = NULL;
+	if ((path = gnome_desktop_item_get_location (ditem)))
+		retval = g_strdup (panel_launcher_get_filename (path));
+
+	gnome_desktop_item_unref (ditem);
+
+	return retval;
+}
 
 static void
 properties_apply (Launcher *launcher)
 {
 	char *location;
+	char *new_location;
 
 	/* save location */
 	location = g_strdup (gnome_desktop_item_get_location (launcher->ditem));
@@ -620,8 +657,25 @@ properties_apply (Launcher *launcher)
 		gnome_ditem_edit_get_ditem (GNOME_DITEM_EDIT (launcher->dedit));
 	launcher->ditem = gnome_desktop_item_copy (launcher->ditem);
 
-	/* restore location */
-	gnome_desktop_item_set_location (launcher->ditem, location);
+	if ((new_location = panel_launcher_hoard (location))) {
+		GConfClient *client;
+		const char  *key;
+
+		client = panel_gconf_get_client ();
+
+		key = panel_gconf_full_key (PANEL_GCONF_OBJECTS,
+					    launcher->info->id,
+					    "launcher_location");
+
+		gconf_client_set_string (client, key, new_location, NULL);
+
+		gnome_desktop_item_set_location (launcher->ditem, new_location);
+
+		g_free (new_location);
+	} else {
+		gnome_desktop_item_set_location (launcher->ditem, location);
+	}
+
 	g_free (location);
 
 	launcher_save (launcher);
@@ -646,9 +700,23 @@ properties_close_callback(GtkWidget *widget, gpointer data)
 }
 
 static void
-launcher_changed (GtkObject *dedit, gpointer data)
+launcher_changed (GObject *dedit, gpointer data)
 {
-	Launcher *launcher = data;
+	Launcher         *launcher;
+	GnomeDesktopItem *new_ditem;
+	const char       *old_exec;
+	const char       *new_exec;
+
+	launcher = (Launcher *) data;
+	new_ditem = gnome_ditem_edit_get_ditem (GNOME_DITEM_EDIT (dedit));
+
+	old_exec = gnome_desktop_item_get_string (launcher->ditem,
+						  GNOME_DESKTOP_ITEM_EXEC);
+	new_exec = gnome_desktop_item_get_string (new_ditem,
+						  GNOME_DESKTOP_ITEM_EXEC);
+
+	if (!old_exec || !new_exec || strcmp (old_exec, new_exec))
+		gnome_desktop_item_clear_attr (new_ditem, "StartupNotify");
 
 	properties_apply (launcher);
 }
@@ -677,6 +745,8 @@ window_response (GtkWidget *w, int response, gpointer data)
 			gtk_window_get_screen (GTK_WINDOW (w)),
 			"user-guide.xml", "gospanel-52");
 	} else if (response == REVERT_BUTTON) { /* revert */
+		char *new_location;
+
 		if (launcher->ditem != NULL)
 			gnome_desktop_item_unref (launcher->ditem);
 		launcher->ditem = gnome_desktop_item_copy (launcher->revert_ditem);
@@ -697,6 +767,23 @@ window_response (GtkWidget *w, int response, gpointer data)
 				  G_CALLBACK (launcher_changed),
 				  launcher);
 
+		if ((new_location = panel_launcher_hoard (gnome_desktop_item_get_location (launcher->ditem)))) {
+			GConfClient *client;
+			const char  *key;
+
+			client = panel_gconf_get_client ();
+
+			key = panel_gconf_full_key (PANEL_GCONF_OBJECTS,
+						    launcher->info->id,
+						    "launcher_location");
+
+			gconf_client_set_string (client, key, new_location, NULL);
+
+			gnome_desktop_item_set_location (launcher->ditem, new_location);
+
+			g_free (new_location);
+		}
+
 		/* resave launcher */
 		launcher_save (launcher);
 
@@ -849,45 +936,6 @@ load_launcher_applet (const char       *location,
 	return launcher;
 }
 
-static void
-panel_launcher_ensure_hoarded (Launcher   *launcher,
-			       const char *launcher_location,
-			       const char *id)
-{
-	GConfClient *client;
-	const char  *key;
-	const char  *new_location = NULL;
-
-	if (!strchr (launcher_location, G_DIR_SEPARATOR))
-		return; /* already hoarded */
-
-	client  = panel_gconf_get_client ();
-	key = panel_gconf_full_key (PANEL_GCONF_OBJECTS, id, "launcher_location");
-	if ( ! gconf_client_key_is_writable (client, key, NULL)) {
-		/* can't hoard */
-		launcher->non_writable = TRUE;
-		return;
-	}
-	
-	if (!(new_location = panel_launcher_get_filename (launcher_location))) {
-		const char *path;
-
-		gnome_desktop_item_set_location (launcher->ditem, NULL);
-		panel_launcher_save_ditem (launcher->ditem, launcher_get_screen (launcher));
-		
-		path = gnome_desktop_item_get_location (launcher->ditem);
-		new_location = panel_launcher_get_filename (path);
-	}
-
-	if (!new_location) {
-		/* hoarding must have failed */
-		launcher->non_writable = TRUE;
-		return;
-	}
-
-	gconf_client_set_string (client, key, new_location, NULL);
-}
-
 void
 launcher_load_from_gconf (PanelWidget *panel_widget,
 			  gboolean     locked,
@@ -905,6 +953,7 @@ launcher_load_from_gconf (PanelWidget *panel_widget,
 	client  = panel_gconf_get_client ();
 
 	key = panel_gconf_full_key (PANEL_GCONF_OBJECTS, id, "launcher_location");
+	panel_compatibility_migrate_applications_scheme (client, key);
 	launcher_location = gconf_client_get_string (client, key, NULL);
 
 	if (!launcher_location) {
@@ -920,10 +969,10 @@ launcher_load_from_gconf (PanelWidget *panel_widget,
 					 id);
 
 	if (launcher) {
-		panel_launcher_ensure_hoarded (launcher, launcher_location, id);
-
-		if (launcher->non_writable) {
+		key = panel_gconf_full_key (PANEL_GCONF_OBJECTS, id, "launcher_location");
+		if (!gconf_client_key_is_writable (client, key, NULL)) {
 			AppletUserMenu *menu;
+
 			menu = panel_applet_get_callback (launcher->info->user_menu,
 							  "properties");
 			if (menu != NULL)
@@ -965,10 +1014,11 @@ really_add_launcher (GtkWidget *dialog, int response, gpointer data)
 							 TRUE,
 						         _("Cannot create launcher"),
 						         _("You have to specify a name."));
-			g_signal_connect_swapped (G_OBJECT (dialog),
-						  "destroy",
-						   G_CALLBACK (gtk_widget_destroy),
-						   G_OBJECT (err_dialog));
+			g_signal_connect_object (G_OBJECT (dialog),
+						 "destroy",
+						 G_CALLBACK (gtk_widget_destroy),
+						 G_OBJECT (err_dialog),
+						 G_CONNECT_SWAPPED);
 			return;
 		}
 		
@@ -983,10 +1033,11 @@ really_add_launcher (GtkWidget *dialog, int response, gpointer data)
 							 TRUE,
 							 _("Cannot create launcher"),
 							 _("You have to specify a valid URL or command."));
-			g_signal_connect_swapped (G_OBJECT (dialog),
-						  "destroy",
-						  G_CALLBACK (gtk_widget_destroy),
-						  G_OBJECT (err_dialog));
+			g_signal_connect_object (G_OBJECT (dialog),
+						 "destroy",
+						 G_CALLBACK (gtk_widget_destroy),
+						 G_OBJECT (err_dialog),
+						 G_CONNECT_SWAPPED);
 			return;
 		}
 
@@ -999,7 +1050,7 @@ really_add_launcher (GtkWidget *dialog, int response, gpointer data)
 		ensure_item_localefiled (ditem, GNOME_DESKTOP_ITEM_GENERIC_NAME);
 		ensure_item_localefiled (ditem, GNOME_DESKTOP_ITEM_COMMENT);
 
-		panel_launcher_save_ditem (ditem, gtk_window_get_screen (GTK_WINDOW (dialog)));
+		panel_launcher_save_ditem (ditem, TRUE, gtk_window_get_screen (GTK_WINDOW (dialog)));
 		location = gnome_desktop_item_get_location (ditem);
 
 		panel_launcher_create (panel->toplevel, pos, location);
@@ -1105,7 +1156,7 @@ panel_launcher_create_from_info (PanelToplevel *toplevel,
 		gnome_desktop_item_set_entry_type (ditem, GNOME_DESKTOP_ITEM_TYPE_LINK);
 	}
 
-	panel_launcher_save_ditem (ditem, gtk_window_get_screen (GTK_WINDOW (toplevel)));
+	panel_launcher_save_ditem (ditem, TRUE, gtk_window_get_screen (GTK_WINDOW (toplevel)));
 	location = gnome_desktop_item_get_location (ditem);
 
 	panel_launcher_create (toplevel, position, location);
@@ -1121,18 +1172,27 @@ panel_launcher_create_with_id (const char    *toplevel_id,
 	GConfClient *client;
 	const char  *key;
 	char        *id;
+	char        *new_location;
 
 	g_return_if_fail (location != NULL);
 
-	client  = panel_gconf_get_client ();
+	client = panel_gconf_get_client ();
 
 	id = panel_profile_prepare_object_with_id (PANEL_OBJECT_LAUNCHER, toplevel_id, position, FALSE);
 
-	key = panel_gconf_full_key (PANEL_GCONF_OBJECTS, id, "launcher_location");
-	gconf_client_set_string (client, key, location, NULL);
+	new_location = panel_launcher_hoard (location);
+
+	key = panel_gconf_full_key (PANEL_GCONF_OBJECTS,
+				    id,
+				    "launcher_location");
+	if (new_location)
+		gconf_client_set_string (client, key, new_location, NULL);
+	else
+		gconf_client_set_string (client, key, location, NULL);
 
 	panel_profile_add_to_list (PANEL_GCONF_OBJECTS, id);
 
+	g_free (new_location);
 	g_free (id);
 }
 
@@ -1182,7 +1242,7 @@ launcher_save (Launcher *launcher)
 	g_return_if_fail (launcher != NULL);
 	g_return_if_fail (launcher->ditem != NULL);
 
-	panel_launcher_save_ditem (launcher->ditem, launcher_get_screen (launcher));
+	panel_launcher_save_ditem (launcher->ditem, TRUE, launcher_get_screen (launcher));
 }
 
 Launcher *
@@ -1230,6 +1290,7 @@ panel_launcher_set_dnd_enabled (Launcher *launcher,
 				     GDK_BUTTON1_MASK,
 				     dnd_targets, 2,
 				     GDK_ACTION_COPY | GDK_ACTION_MOVE);
+		//FIXME: this doesn't work since the pixbuf isn't loaded yet
 		if (BUTTON_WIDGET (launcher->button)->pixbuf)
 			gtk_drag_source_set_icon_pixbuf (launcher->button,
 							 BUTTON_WIDGET (launcher->button)->pixbuf);
