@@ -39,6 +39,7 @@
 #include "panel-marshal.h"
 #include "panel-background.h"
 #include "panel-lockdown.h"
+#include "panel-stock-icons.h"
 
 #define PANEL_APPLET_FRAME_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), PANEL_TYPE_APPLET_FRAME, PanelAppletFramePrivate))
 
@@ -46,8 +47,6 @@
 
 #define HANDLE_SIZE 10
 
-#define PANEL_STOCK_DONT_DELETE "panel-dont-delete"
-
 #define PROPERTY_ORIENT     "panel-applet-orient"
 #define PROPERTY_SIZE       "panel-applet-size"
 #define PROPERTY_BACKGROUND "panel-applet-background"
@@ -140,7 +139,7 @@ panel_applet_frame_sync_menu_state (PanelAppletFrame *frame)
 	bonobo_ui_component_set_prop (frame->priv->ui_component,
 				      "/commands/RemoveAppletFromPanel",
 				      "sensitive",
-				      removable ? "1" : "0",
+				      (locked && !lockable) ? "0" : (removable ? "1" : "0"),
 				      NULL);
 
 	bonobo_ui_component_set_prop (frame->priv->ui_component,
@@ -407,16 +406,16 @@ panel_applet_frame_change_orientation (PanelAppletFrame *frame,
 
 	switch (orientation) {
 	case PANEL_ORIENTATION_TOP:
-		orient = GNOME_Vertigo_PANEL_ORIENT_DOWN;
-		break;
-	case PANEL_ORIENTATION_BOTTOM:
 		orient = GNOME_Vertigo_PANEL_ORIENT_UP;
 		break;
+	case PANEL_ORIENTATION_BOTTOM:
+		orient = GNOME_Vertigo_PANEL_ORIENT_DOWN;
+		break;
 	case PANEL_ORIENTATION_LEFT:
-		orient = GNOME_Vertigo_PANEL_ORIENT_RIGHT;
+		orient = GNOME_Vertigo_PANEL_ORIENT_LEFT;
 		break;
 	case PANEL_ORIENTATION_RIGHT:
-		orient = GNOME_Vertigo_PANEL_ORIENT_LEFT;
+		orient = GNOME_Vertigo_PANEL_ORIENT_RIGHT;
 		break;
 	default:
 		g_assert_not_reached ();
@@ -963,33 +962,6 @@ panel_applet_frame_cnx_broken (ORBitConnection  *cnx,
 	g_free (dialog_txt);
 }
 
-static inline void
-register_stock_item (void)
-{
-        static gboolean registered = FALSE;
-
-        if (!registered) {
-                GtkIconFactory      *factory;
-                GtkIconSet          *cancel_icons;
-
-                static GtkStockItem  dont_delete_item [] = {
-                        { PANEL_STOCK_DONT_DELETE, N_("D_on't Delete"), 0, 0, GETTEXT_PACKAGE },
-                };
-
-                cancel_icons = gtk_icon_factory_lookup_default (GTK_STOCK_CANCEL);
-
-                factory = gtk_icon_factory_new ();
-
-                gtk_icon_factory_add (factory, PANEL_STOCK_DONT_DELETE, cancel_icons);
-
-                gtk_icon_factory_add_default (factory);
-
-                gtk_stock_add_static (dont_delete_item, 1);
-
-                registered = TRUE;
-        }
-}
-
 enum {
 	LOADING_FAILED_RESPONSE_DONT_DELETE,
 	LOADING_FAILED_RESPONSE_DELETE
@@ -1019,7 +991,6 @@ panel_applet_frame_loading_failed (PanelAppletFrame  *frame,
 	gboolean   locked_down;
 
 	locked_down = panel_lockdown_get_locked_down ();
-	register_stock_item ();
 
 	problem_txt = g_strdup_printf (_("The panel encountered a problem "
 					 "while loading \"%s\"."),
@@ -1048,7 +1019,7 @@ panel_applet_frame_loading_failed (PanelAppletFrame  *frame,
 	}
 
 	gtk_window_set_screen (GTK_WINDOW (dialog),
-			       gtk_window_get_screen (GTK_WINDOW (frame->priv->panel)));
+			       gtk_window_get_screen (GTK_WINDOW (frame->priv->panel->toplevel)));
 
 	g_signal_connect (dialog, "response",
 			  G_CALLBACK (panel_applet_frame_loading_failed_response),
@@ -1156,16 +1127,16 @@ panel_applet_frame_get_orient_string (PanelAppletFrame *frame,
 
 	switch (orientation) {
 	case PANEL_ORIENTATION_TOP:
-		retval = "down";
-		break;
-	case PANEL_ORIENTATION_BOTTOM:
 		retval = "up";
 		break;
+	case PANEL_ORIENTATION_BOTTOM:
+		retval = "down";
+		break;
 	case PANEL_ORIENTATION_LEFT:
-		retval = "right";
+		retval = "left";
 		break;
 	case PANEL_ORIENTATION_RIGHT:
-		retval = "left";
+		retval = "right";
 		break;
 	default:
 		g_assert_not_reached ();
@@ -1275,9 +1246,6 @@ panel_applet_frame_activated (Bonobo_Unknown     object,
 	}
 
 	frame->priv->control = CORBA_Object_duplicate (object, NULL);
-	ORBit_small_listen_for_broken (object,
-				       G_CALLBACK (panel_applet_frame_cnx_broken),
-				       frame);
 
 	widget = bonobo_widget_new_control_from_objref (object,
 							CORBA_OBJECT_NIL);
@@ -1381,6 +1349,10 @@ panel_applet_frame_activated (Bonobo_Unknown     object,
 
 	CORBA_exception_free (&corba_ev);
 
+	ORBit_small_listen_for_broken (object,
+				       G_CALLBACK (panel_applet_frame_cnx_broken),
+				       frame);
+
 	gtk_container_add (GTK_CONTAINER (frame), widget);
 
 	gtk_widget_show_all (GTK_WIDGET (frame));
